@@ -214,7 +214,7 @@ class Buffer:
 
     `Server`, `RolloutWorker` and `Trainer` all have their own `Buffer` to store and send training samples.
 
-    Samples are tuples of the form (`act`, `new_obs`, `rew`, `done`, `info`)
+    Samples are tuples of the form (`act`, `new_obs`, `rew`, `terminated`, `truncated`, `info`)
     """
     def __init__(self, maxlen=cfg.BUFFERS_MAXLEN):
         """
@@ -239,7 +239,7 @@ class Buffer:
         Appends `sample` to the buffer.
 
         Args:
-            sample (Tuple): a training sample of the form (`act`, `new_obs`, `rew`, `done`, `info`)
+            sample (Tuple): a training sample of the form (`act`, `new_obs`, `rew`, `terminated`, `truncated`, `info`)
         """
         self.memory.append(sample)
         self.clip_to_maxlen()
@@ -558,7 +558,8 @@ def iterate_epochs_tm(run_cls,
                       checkpoint_path: str,
                       dump_run_instance_fn=dump_run_instance,
                       load_run_instance_fn=load_run_instance,
-                      epochs_between_checkpoints=1):
+                      epochs_between_checkpoints=1,
+                      updater_fn=None):
     """
     Main training loop (remote)
     The run_cls instance is saved in checkpoint_path at the end of each epoch
@@ -568,17 +569,22 @@ def iterate_epochs_tm(run_cls,
     checkpoint_path = checkpoint_path or tempfile.mktemp("_remove_on_exit")
 
     try:
-        print(f"DEBUF: checkpoint_path: {checkpoint_path}")
+        logging.debug(f"checkpoint_path: {checkpoint_path}")
         if not exists(checkpoint_path):
             logging.info(f"=== specification ".ljust(70, "="))
             run_instance = run_cls()
             dump_run_instance_fn(run_instance, checkpoint_path)
             logging.info(f"")
         else:
-            logging.info(f" Loading checkpoint...")
+            logging.info(f"Loading checkpoint...")
             t1 = time.time()
             run_instance = load_run_instance_fn(checkpoint_path)
             logging.info(f" Loaded checkpoint in {time.time() - t1} seconds.")
+            if updater_fn is not None:
+                logging.info(f"Updating checkpoint...")
+                t1 = time.time()
+                run_instance = updater_fn(run_instance, run_cls)
+                logging.info(f"Checkpoint updated in {time.time() - t1} seconds.")
 
         while run_instance.epoch < run_instance.epochs:
             # time.sleep(1)  # on network file systems writing files is asynchronous and we need to wait for sync
@@ -598,7 +604,7 @@ def iterate_epochs_tm(run_cls,
             os.remove(checkpoint_path)
 
 
-def run_with_wandb(entity, project, run_id, interface, run_cls, checkpoint_path: str = None, dump_run_instance_fn=None, load_run_instance_fn=None):
+def run_with_wandb(entity, project, run_id, interface, run_cls, checkpoint_path: str = None, dump_run_instance_fn=None, load_run_instance_fn=None, updater_fn=None):
     """
     Main training loop (remote).
 
@@ -614,19 +620,32 @@ def run_with_wandb(entity, project, run_id, interface, run_cls, checkpoint_path:
     config['environ'] = log_environment_variables()
     # config['git'] = git_info()  # TODO: check this for bugs
     resume = checkpoint_path and exists(checkpoint_path)
-    wandb.init(dir=wandb_dir, entity=entity, project=project, id=run_id, resume=resume, config=config)
+    wandb_initialized = False
+    err_cpt = 0
+    while not wandb_initialized:
+        try:
+            wandb.init(dir=wandb_dir, entity=entity, project=project, id=run_id, resume=resume, config=config)
+            wandb_initialized = True
+        except Exception as e:
+            err_cpt += 1
+            logging.warning(f"wandb error {err_cpt}: {e}")
+            if err_cpt > 10:
+                logging.warning(f"Could not connect to wandb, aborting.")
+                exit()
+            else:
+                time.sleep(10.0)
     # logging.info(config)
-    for stats in iterate_epochs_tm(run_cls, interface, checkpoint_path, dump_run_instance_fn, load_run_instance_fn):
+    for stats in iterate_epochs_tm(run_cls, interface, checkpoint_path, dump_run_instance_fn, load_run_instance_fn, 1, updater_fn):
         [wandb.log(json.loads(s.to_json())) for s in stats]
 
 
-def run(interface, run_cls, checkpoint_path: str = None, dump_run_instance_fn=None, load_run_instance_fn=None):
+def run(interface, run_cls, checkpoint_path: str = None, dump_run_instance_fn=None, load_run_instance_fn=None, updater_fn=None):
     """
     Main training loop (remote).
     """
     dump_run_instance_fn = dump_run_instance_fn or dump_run_instance
     load_run_instance_fn = load_run_instance_fn or load_run_instance
-    for stats in iterate_epochs_tm(run_cls, interface, checkpoint_path, dump_run_instance_fn, load_run_instance_fn):
+    for stats in iterate_epochs_tm(run_cls, interface, checkpoint_path, dump_run_instance_fn, load_run_instance_fn, 1, updater_fn):
         pass
 
 
@@ -643,7 +662,8 @@ class Trainer:
                  model_path=cfg.MODEL_PATH_TRAINER,
                  checkpoint_path=cfg.CHECKPOINT_PATH,
                  dump_run_instance_fn: callable = None,
-                 load_run_instance_fn: callable = None):
+                 load_run_instance_fn: callable = None,
+                 updater_fn: callable = None):
         """
         Args:
             training_cls (type): training class (subclass of tmrl.training_offline.TrainingOffline)
@@ -652,10 +672,14 @@ class Trainer:
             checkpoint_path: path where the `Trainer` will be checkpointed (`None` = no checkpointing)
             dump_run_instance_fn (callable): custom serializer (`None` = pickle.dump)
             load_run_instance_fn (callable): custom deserializer (`None` = pickle.load)
+            updater_fn (callable): custom updater (`None` = no updater). If provided, this must be a function \
+            that takes a checkpoint and training_cls as argument and returns an updated checkpoint. \
+            The updater is called after a checkpoint is loaded, e.g., to update your checkpoint with new arguments.
         """
         self.checkpoint_path = checkpoint_path
         self.dump_run_instance_fn = dump_run_instance_fn
         self.load_run_instance_fn = load_run_instance_fn
+        self.updater_fn = updater_fn
         self.training_cls = training_cls
         self.interface = TrainerInterface(server_ip=server_ip,
                                           model_path=model_path)
@@ -668,7 +692,8 @@ class Trainer:
             run_cls=self.training_cls,
             checkpoint_path=self.checkpoint_path,
             dump_run_instance_fn=self.dump_run_instance_fn,
-            load_run_instance_fn=self.load_run_instance_fn)
+            load_run_instance_fn=self.load_run_instance_fn,
+            updater_fn=self.updater_fn)
 
     def run_with_wandb(self,
                        entity=cfg.WANDB_ENTITY,
@@ -695,7 +720,8 @@ class Trainer:
                        run_cls=self.training_cls,
                        checkpoint_path=self.checkpoint_path,
                        dump_run_instance_fn=self.dump_run_instance_fn,
-                       load_run_instance_fn=self.load_run_instance_fn)
+                       load_run_instance_fn=self.load_run_instance_fn,
+                       updater_fn=self.updater_fn)
 
 
 # ROLLOUT WORKER: ===================================
@@ -734,7 +760,7 @@ class RolloutWorker:
             min_samples_per_worker_packet (int): the worker waits for this number of samples before sending
             max_samples_per_episode (int): if an episode gets longer than this, it is reset
             model_path (str): path where a local copy of the policy will be stored
-            obs_preprocessor (callable): utility for modifying samples before forward passes
+            obs_preprocessor (callable): utility for modifying observations retrieved from the environment
             crc_debug (bool): can be used for debugging the pipeline
             model_path_history (str): (omit .pth) an history of policies can be stored here
             model_history (int): new policies are saved % model_history (0: not saved)
@@ -861,30 +887,30 @@ class RolloutWorker:
 
         Returns:
             obs (nested structure): observation retrieved from the environment
+            info (dict): information retrieved from the environment
         """
         obs = None
         act = self.env.default_action.astype(np.float32)
-        new_obs = self.env.reset()
+        new_obs, info = self.env.reset()
         if self.obs_preprocessor is not None:
             new_obs = self.obs_preprocessor(new_obs)
         rew = 0.0
-        done = False
-        info = {}
+        terminated, truncated = False, False
         if collect_samples:
             if self.crc_debug:
-                info['crc_sample'] = (obs, act, new_obs, rew, done)
+                info['crc_sample'] = (obs, act, new_obs, rew, terminated, truncated)
             if self.get_local_buffer_sample:
-                sample = self.get_local_buffer_sample(act, new_obs, rew, done, info)
+                sample = self.get_local_buffer_sample(act, new_obs, rew, terminated, truncated, info)
             else:
-                sample = act, new_obs, rew, done, info
+                sample = act, new_obs, rew, terminated, truncated, info
             self.buffer.append_sample(sample)
-        return new_obs
+        return new_obs, info
 
     def step(self, obs, test, collect_samples, last_step=False):
         """
         Performs a full RL transition.
 
-        A full RL transition is `obs` -> `act` -> `new_obs`, `rew`, `done`, `info`.
+        A full RL transition is `obs` -> `act` -> `new_obs`, `rew`, `terminated`, `truncated`, `info`.
         Note that, in the Real-Time RL setting, `act` is appended to a buffer which is part of `new_obs`.
         This is because is does not directly affect the new observation, due to real-time delays.
 
@@ -892,52 +918,50 @@ class RolloutWorker:
             obs (nested structure): previous observation
             test (bool): passed to the `act()` method of the `ActorModule`
             collect_samples (bool): if True, samples are buffered and sent to the `Server`
-            last_step (bool): if True and `done` is False, a '__no_done' entry will be added to the `info` dict
+            last_step (bool): if True and `terminated` is False, `truncated` will be set to True
 
         Returns:
             new_obs (nested structure): new observation
             rew (float): new reward
-            done (bool): episode termination signal
+            terminated (bool): episode termination signal
+            truncated (bool): episode truncation signal
             info (dict): information dictionary
         """
         act = self.act(obs, test=test)
-        new_obs, rew, done, info = self.env.step(act)
+        new_obs, rew, terminated, truncated, info = self.env.step(act)
         if self.obs_preprocessor is not None:
             new_obs = self.obs_preprocessor(new_obs)
         if collect_samples:
-            stored_done = done
-            if last_step and not done:  # ignore done when stopped by step limit
-                info["__no_done"] = True
-            if "__no_done" in info:
-                stored_done = False
+            if last_step and not terminated:
+                truncated = True
             if self.crc_debug:
-                info['crc_sample'] = (obs, act, new_obs, rew, stored_done)
+                info['crc_sample'] = (obs, act, new_obs, rew, terminated, truncated)
             if self.get_local_buffer_sample:
-                sample = self.get_local_buffer_sample(act, new_obs, rew, stored_done, info)
+                sample = self.get_local_buffer_sample(act, new_obs, rew, terminated, truncated, info)
             else:
-                sample = act, new_obs, rew, stored_done, info
+                sample = act, new_obs, rew, terminated, truncated, info
             self.buffer.append_sample(sample)  # CAUTION: in the buffer, act is for the PREVIOUS transition (act, obs(act))
-        return new_obs, rew, done, info
+        return new_obs, rew, terminated, truncated, info
 
     def collect_train_episode(self, max_samples):
         """
-        Collects a maximum of n training transitions (from reset to done)
+        Collects a maximum of n training transitions (from reset to terminated or truncated)
 
         This method stores the episode and the train return in the local `Buffer` of the worker
         for sending to the `Server`.
 
         Args:
-            max_samples (int): if the environment is not `done` after `max_samples` time steps,
-                it is forcefully reset and a '__no_done' entry is added to the `info` dict of the corresponding sample.
+            max_samples (int): if the environment is not `terminated` after `max_samples` time steps,
+                it is forcefully reset and `truncated` is set to True.
         """
         ret = 0.0
         steps = 0
-        obs = self.reset(collect_samples=True)
+        obs, info = self.reset(collect_samples=True)
         for i in range(max_samples):
-            obs, rew, done, info = self.step(obs=obs, test=False, collect_samples=True, last_step=i == max_samples - 1)
+            obs, rew, terminated, truncated, info = self.step(obs=obs, test=False, collect_samples=True, last_step=i == max_samples - 1)
             ret += rew
             steps += 1
-            if done:
+            if terminated or truncated:
                 break
         self.buffer.stat_train_return = ret
         self.buffer.stat_train_steps = steps
@@ -958,28 +982,27 @@ class RolloutWorker:
 
     def run_episode(self, max_samples, train=False):
         """
-        Collects a maximum of n test transitions (from reset to done).
+        Collects a maximum of n test transitions (from reset to terminated or truncated).
 
         Args:
             max_samples (int): At most `max_samples` samples are collected per episode.
-                If the episode is longer, it is forcefully reset and a '__no_done' entry is added to the `info` dict
-                of the corresponding sample.
+                If the episode is longer, it is forcefully reset and `truncated` is set to True.
             train (bool): whether the episode is a training or a test episode.
                 `step` is called with `test=not train`.
         """
         ret = 0.0
         steps = 0
-        obs = self.reset(collect_samples=False)
+        obs, info = self.reset(collect_samples=False)
         for _ in range(max_samples):
-            obs, rew, done, info = self.step(obs=obs, test=not train, collect_samples=False)
+            obs, rew, terminated, truncated, info = self.step(obs=obs, test=not train, collect_samples=False)
             ret += rew
             steps += 1
-            if done:
+            if terminated or truncated:
                 break
         self.buffer.stat_test_return = ret
         self.buffer.stat_test_steps = steps
 
-    def run(self, test_episode_interval=20, nb_episodes=np.inf):  # TODO: check number of collected samples are collected before sending
+    def run(self, test_episode_interval=50, nb_episodes=np.inf):  # TODO: check number of collected samples are collected before sending
         """
         Runs the worker for `nb_episodes` episodes.
 
@@ -1008,7 +1031,7 @@ class RolloutWorker:
 
     def profile_step(self):
         import torch.autograd.profiler as profiler
-        obs = self.reset(collect_samples=True)
+        obs, info = self.reset(collect_samples=True)
         use_cuda = True if self.device == 'cuda' else False
         print_with_timestamp(f"use_cuda:{use_cuda}")
         with profiler.profile(record_shapes=True, use_cuda=use_cuda) as prof:
@@ -1032,11 +1055,11 @@ class RolloutWorker:
             nb_steps (int): number of steps to perform to compute the benchmark
             test (int): whether the actor is called in test or train mode
         """
-        obs = self.reset(collect_samples=False)
+        obs, info = self.reset(collect_samples=False)
         for _ in range(nb_steps):
-            obs, rew, done, info = self.step(obs=obs, test=test, collect_samples=False)
-            if done:
-                obs = self.reset(collect_samples=False)
+            obs, rew, terminated, truncated, info = self.step(obs=obs, test=test, collect_samples=False)
+            if terminated or truncated:
+                obs, info = self.reset(collect_samples=False)
         print_with_timestamp(f"Benchmark results:\n{self.env.benchmarks()}")
 
     def send_and_clear_buffer(self):
